from __future__ import annotations

import abc
import logging
import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Text, Tuple

import sqlalchemy as sa
import sqlalchemy.orm
from packaging import version
from sqlalchemy.engine import URL
from sqlalchemy.exc import OperationalError
from typing_extensions import Protocol, runtime_checkable

import rasa
import rasa.model
import rasa.shared.utils.common
import rasa.utils.common
from rasa.constants import MINIMUM_COMPATIBLE_VERSION
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.engine.caching import (
    get_cache_database_name,
    get_local_cache_location,
    get_max_cache_size,
)

logger = logging.getLogger(__name__)


class TrainingCache(abc.ABC):
    """Stores training results in a persistent cache.

    Used to minimize re-retraining when the data / config didn't change in between
    training runs.
    """

    @abc.abstractmethod
    def cache_output(
        self,
        fingerprint_key: Text,
        output: Any,
        output_fingerprint: Text,
        model_storage: ModelStorage,
    ) -> None:
        """Adds the output to the cache.

        If the output is of type `Cacheable` the output is persisted to disk in addition
        to its fingerprint.

        Args:
            fingerprint_key: The fingerprint key serves as key for the cache. Graph
                components can use their fingerprint key to lookup fingerprints of
                previous training runs.
            output: The output. The output is only cached to disk if it's of type
                `Cacheable`.
            output_fingerprint: The fingerprint of their output. This can be used
                to lookup potentially persisted outputs on disk.
            model_storage: Required for caching `Resource` instances. E.g. `Resource`s
                use that to copy data from the model storage to the cache.
        """

    ...

    @abc.abstractmethod
    def get_cached_output_fingerprint(self, fingerprint_key: Text) -> Optional[Text]:
        """Retrieves fingerprint of output based on fingerprint key.

        Args:
            fingerprint_key: The fingerprint serves as key for the lookup of output
                fingerprints.

        Returns:
            The fingerprint of a matching output or `None` in case no cache entry was
            found for the given fingerprint key.
        """
        ...

    @abc.abstractmethod
    def get_cached_result(
        self, output_fingerprint_key: Text, node_name: Text, model_storage: ModelStorage
    ) -> Optional[Cacheable]:
        """Returns a potentially cached output result.

        Args:
            output_fingerprint_key: The fingerprint key of the output serves as lookup
                key for a potentially cached version of this output.
            node_name: The name of the graph node which wants to use this cached result.
            model_storage: The current model storage (e.g. used when restoring
                `Resource` objects so that they can fill the model storage with data).

        Returns:
            `None` if no matching result was found or restored `Cacheable`.
        """
        ...


@runtime_checkable
class Cacheable(Protocol):
    """Protocol for cacheable graph component outputs.

    We only cache graph component outputs which are `Cacheable`. We only store the
    output fingerprint for everything else.
    """

    def to_cache(self, directory: Path, model_storage: ModelStorage) -> None:
        """Persists `Cacheable` to disk.

        Args:
            directory: The directory where the `Cacheable` can persist itself to.
            model_storage: The current model storage (e.g. used when caching `Resource`
                objects.
        """
        ...

    @classmethod
    def from_cache(
        cls,
        node_name: Text,
        directory: Path,
        model_storage: ModelStorage,
        output_fingerprint: Text,
    ) -> Cacheable:
        """Loads `Cacheable` from cache.

        Args:
            node_name: The name of the graph node which wants to use this cached result.
            directory: Directory containing the persisted `Cacheable`.
            model_storage: The current model storage (e.g. used when restoring
                `Resource` objects so that they can fill the model storage with data).
            output_fingerprint: The fingerprint of the cached result (e.g. used when
                restoring `Resource` objects as the fingerprint can not be easily
                calculated from the object itself).

        Returns:
            Instantiated `Cacheable`.
        """
        ...


class LocalTrainingCache(TrainingCache):
    """Caches training results on local disk (see parent class for full docstring)."""

    from sqlalchemy.orm import DeclarativeBase

    class Base(DeclarativeBase):
        pass

    class CacheEntry(Base):
        """Stores metadata about a single cache entry."""

        __tablename__ = "cache_entry"

        fingerprint_key = sa.Column(sa.String(), primary_key=True)
        output_fingerprint_key = sa.Column(sa.String(), nullable=False, index=True)
        last_used = sa.Column(sa.DateTime(timezone=True), nullable=False)
        rasa_version = sa.Column(sa.String(255), nullable=False)
        result_location = sa.Column(sa.String())
        result_type = sa.Column(sa.String())

    def __init__(self) -> None:
        """Creates cache.

        The `Cache` setting can be configured via environment variables.
        """
        self._cache_location = LocalTrainingCache._get_cache_location()

        self._max_cache_size = get_max_cache_size()

        self._cache_database_name = get_cache_database_name()

        if not self._cache_location.exists() and not self._is_disabled():
            logger.debug(
                f"Creating caching directory '{self._cache_location}' because "
                f"it doesn't exist yet."
            )
            self._cache_location.mkdir(parents=True)

        self._sessionmaker = self._create_database()

        self._drop_cache_entries_from_incompatible_versions()

    @staticmethod
    def _get_cache_location() -> Path:
        return get_local_cache_location()

    def _create_database(self) -> sqlalchemy.orm.sessionmaker:
        if self._is_disabled():
            # Use in-memory database as mock to avoid having to check `_is_disabled`
            # everywhere
            database = ""
        else:
            database = str(self._cache_location / self._cache_database_name)

        # Use `future=True` as we are using the 2.x query style
        engine = sa.create_engine(
            URL.create(drivername="sqlite", database=database), future=True
        )
        self.Base.metadata.create_all(engine)

        return sa.orm.sessionmaker(engine)

    def _drop_cache_entries_from_incompatible_versions(self) -> None:
        incompatible_entries = self._find_incompatible_cache_entries()

        for entry in incompatible_entries:
            self._delete_cached_result(entry)

        self._delete_incompatible_entries_from_cache(incompatible_entries)

        logger.debug(
            f"Deleted {len(incompatible_entries)} from disk as their version "
            f"is older than the minimum compatible version "
            f"('{MINIMUM_COMPATIBLE_VERSION}')."
        )

    def _find_incompatible_cache_entries(self) -> List[LocalTrainingCache.CacheEntry]:
        with self._sessionmaker() as session:
            query_for_cache_entries = sa.select(self.CacheEntry)
            all_entries: List[LocalTrainingCache.CacheEntry] = (
                session.execute(query_for_cache_entries).scalars().all()
            )

        return [
            entry
            for entry in all_entries
            if version.parse(MINIMUM_COMPATIBLE_VERSION)
            > version.parse(entry.rasa_version)
        ]

    def _delete_incompatible_entries_from_cache(
        self, incompatible_entries: List[LocalTrainingCache.CacheEntry]
    ) -> None:
        incompatible_fingerprints = [
            entry.fingerprint_key for entry in incompatible_entries
        ]
        with self._sessionmaker.begin() as session:
            delete_query = sa.delete(self.CacheEntry).where(
                self.CacheEntry.fingerprint_key.in_(incompatible_fingerprints)
            )
            session.execute(delete_query)

    @staticmethod
    def _delete_cached_result(entry: LocalTrainingCache.CacheEntry) -> None:
        if entry.result_location and Path(entry.result_location).is_dir():
            shutil.rmtree(entry.result_location)

    def cache_output(
        self,
        fingerprint_key: Text,
        output: Any,
        output_fingerprint: Text,
        model_storage: ModelStorage,
    ) -> None:
        """Adds the output to the cache (see parent class for full docstring)."""
        if self._is_disabled():
            return

        cache_dir, output_type = None, None
        if isinstance(output, Cacheable):
            cache_dir, output_type = self._cache_output_to_disk(output, model_storage)

        try:
            self._add_cache_entry(
                cache_dir, fingerprint_key, output_fingerprint, output_type
            )
        except OperationalError:
            if cache_dir:
                shutil.rmtree(cache_dir)

            raise

    def _add_cache_entry(
        self,
        cache_dir: Optional[Text],
        fingerprint_key: Text,
        output_fingerprint: Text,
        output_type: Text,
    ) -> None:
        with self._sessionmaker.begin() as session:
            cache_entry = self.CacheEntry(
                fingerprint_key=fingerprint_key,
                output_fingerprint_key=output_fingerprint,
                last_used=datetime.utcnow(),
                rasa_version=rasa.__version__,
                result_location=cache_dir,
                result_type=output_type,
            )
            session.merge(cache_entry)

    def _is_disabled(self) -> bool:
        return self._max_cache_size == 0.0

    def _cache_output_to_disk(
        self, output: Cacheable, model_storage: ModelStorage
    ) -> Tuple[Optional[Text], Optional[Text]]:
        tempdir_name = rasa.utils.common.get_temp_dir_name()

        # Use `TempDirectoryPath` instead of `tempfile.TemporaryDirectory` as this
        # leads to errors on Windows when the context manager tries to delete an
        # already deleted temporary directory (e.g. https://bugs.python.org/issue29982)
        with rasa.utils.common.TempDirectoryPath(tempdir_name) as temp_dir:
            tmp_path = Path(temp_dir)
            try:
                output.to_cache(tmp_path, model_storage)

                logger.debug(
                    f"Caching output of type '{type(output).__name__}' succeeded."
                )
            except Exception as e:
                logger.error(
                    f"Caching output of type '{type(output).__name__}' failed with the "
                    f"following error:\n{e}"
                )
                return None, None

            output_size = rasa.utils.common.directory_size_in_mb(tmp_path)
            if output_size > self._max_cache_size:
                logger.debug(
                    f"Caching result of type '{type(output).__name__}' was skipped "
                    f"because it exceeds the maximum cache size of "
                    f"{self._max_cache_size} MiB."
                )
                return None, None

            while (
                rasa.utils.common.directory_size_in_mb(
                    self._cache_location,
                    filenames_to_exclude=[self._cache_database_name],
                )
                + output_size
                > self._max_cache_size
            ):
                self._drop_least_recently_used_item()

            output_type = rasa.shared.utils.common.module_path_from_instance(output)
            cache_path = shutil.move(temp_dir, self._cache_location)

            return cache_path, output_type

    def _drop_least_recently_used_item(self) -> None:
        with self._sessionmaker.begin() as session:
            query_for_least_recently_used_entry = sa.select(self.CacheEntry).order_by(
                self.CacheEntry.last_used.asc()
            )
            oldest_cache_item = (
                session.execute(query_for_least_recently_used_entry).scalars().first()
            )

            if not oldest_cache_item:
                self._purge_cache_dir_content()
                return

            self._delete_cached_result(oldest_cache_item)
            delete_query = sa.delete(self.CacheEntry).where(
                self.CacheEntry.fingerprint_key == oldest_cache_item.fingerprint_key
            )
            session.execute(delete_query)

            logger.debug(
                f"Deleted item with fingerprint "
                f"'{oldest_cache_item.fingerprint_key}' to free space."
            )

    def _purge_cache_dir_content(self) -> None:
        for item in self._cache_location.glob("*"):
            if item.name == self._cache_database_name:
                continue

            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()

    def get_cached_output_fingerprint(self, fingerprint_key: Text) -> Optional[Text]:
        """Returns cached output fingerprint (see parent class for full docstring)."""
        with self._sessionmaker.begin() as session:
            query = sa.select(self.CacheEntry).filter_by(
                fingerprint_key=fingerprint_key
            )
            match = session.execute(query).scalars().first()

            if match:
                # This result was used during a fingerprint run.
                match.last_used = datetime.utcnow()
                return match.output_fingerprint_key

            return None

    def get_cached_result(
        self, output_fingerprint_key: Text, node_name: Text, model_storage: ModelStorage
    ) -> Optional[Cacheable]:
        """Returns a potentially cached output (see parent class for full docstring)."""
        result_location, result_type = self._get_cached_result(output_fingerprint_key)

        if not result_location:
            logger.debug(f"No cached output found for '{output_fingerprint_key}'")
            return None

        path_to_cached = Path(result_location)
        if not path_to_cached.is_dir():
            logger.debug(
                f"Cached output for '{output_fingerprint_key}' can't be found on disk."
            )
            return None

        return self._load_from_cache(
            result_location,
            result_type,
            node_name,
            model_storage,
            output_fingerprint_key,
        )

    def _get_cached_result(
        self, output_fingerprint_key: Text
    ) -> Tuple[Optional[Path], Optional[Text]]:
        with self._sessionmaker.begin() as session:
            query = sa.select(
                self.CacheEntry.result_location, self.CacheEntry.result_type
            ).where(
                self.CacheEntry.output_fingerprint_key == output_fingerprint_key,
                self.CacheEntry.result_location != sa.null(),
            )

            match = session.execute(query).first()

            if match:
                return Path(match.result_location), match.result_type

            return None, None

    @staticmethod
    def _load_from_cache(
        path_to_cached: Path,
        result_type: Text,
        node_name: Text,
        model_storage: ModelStorage,
        output_fingerprint_key: Text,
    ) -> Optional[Cacheable]:
        try:
            module = rasa.shared.utils.common.class_from_module_path(result_type)

            if not isinstance(module, Cacheable):
                logger.warning(
                    "Failed to restore a non cacheable module from cache. "
                    "Please implement the 'Cacheable' interface for module "
                    f"'{result_type}'."
                )
                return None

            return module.from_cache(
                node_name, path_to_cached, model_storage, output_fingerprint_key
            )
        except Exception as e:
            logger.warning(
                f"Failed to restore cached output of type '{result_type}' from "
                f"cache. Error:\n{e}"
            )
            return None



from __future__ import annotations

from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    OPENAI_PROVIDER,
    RASA_PROVIDER,
)
from rasa.shared.providers._configs.rasa_llm_client_config import (
    RasaLLMClientConfig,
)
from rasa.shared.providers.constants import (
    LITE_LLM_API_BASE_FIELD,
    LITE_LLM_API_KEY_FIELD,
)
from rasa.shared.providers.llm._base_litellm_client import _BaseLiteLLMClient
from rasa.utils.licensing import retrieve_license_from_env

structlogger = structlog.get_logger()


class RasaLLMClient(_BaseLiteLLMClient):
    """A client for interfacing with a Rasa-Hosted LLM endpoint that uses.

    Parameters:
        model (str): The model or deployment name.
        api_base (str): The base URL of the API endpoint.
        kwargs: Any: Additional configuration parameters that can include, but
            are not limited to model parameters and lite-llm specific
            parameters. These parameters will be passed to the
            completion/acompletion calls. To see what it can include, visit:

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
        ProviderClientAPIException: If the API request fails.
    """

    def __init__(
        self,
        model: str,
        api_base: str,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._model = model
        self._api_base = api_base
        self._use_chat_completions_endpoint = True
        self._extra_parameters = kwargs or {}

    @property
    def model(self) -> str:
        return self._model

    @property
    def api_base(self) -> Optional[str]:
        """Returns the base API URL for the openai llm client."""
        return self._api_base

    @property
    def provider(self) -> str:
        """Returns the provider name for the self hosted llm client.

        Returns:
            String representing the provider name.
        """
        return RASA_PROVIDER

    @property
    def _litellm_model_name(self) -> str:
        return f"{OPENAI_PROVIDER}/{self._model}"

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        return self._extra_parameters

    @property
    def config(self) -> dict:
        return RasaLLMClientConfig(
            model=self._model,
            api_base=self._api_base,
            extra_parameters=self._extra_parameters,
        ).to_dict()

    @property
    def _completion_fn_args(self) -> Dict[str, Any]:
        """Returns the completion arguments for invoking a call using completions."""
        fn_args = super()._completion_fn_args
        fn_args.update(
            {
                LITE_LLM_API_BASE_FIELD: self.api_base,
                LITE_LLM_API_KEY_FIELD: retrieve_license_from_env(),
            }
        )
        return fn_args

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> RasaLLMClient:
        try:
            client_config = RasaLLMClientConfig.from_dict(config)
        except ValueError as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "rasa_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise
        return cls(
            model=client_config.model,
            api_base=client_config.api_base,
            **client_config.extra_parameters,
        )



from __future__ import annotations

import dataclasses
from typing import Any, Dict, List, Optional, Union

from rasa.shared.constants import DEFAULT_ENDPOINTS_PATH
from rasa.shared.core.constants import (
    GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE,
    GLOBAL_SILENCE_TIMEOUT_KEY,
)
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import (
    EndpointConfig,
    read_endpoint_config,
    read_property_config_from_endpoints_file,
)


@dataclasses.dataclass
class InteractionHandlingConfig:
    """Configuration for interaction handling."""

    global_silence_timeout: Union[float, int] = GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE

    def __post_init__(self) -> None:
        # Validate the type of `global_silence_timeout`.
        if isinstance(self.global_silence_timeout, str):
            try:
                self.global_silence_timeout = float(self.global_silence_timeout)
            except ValueError:
                raise RasaException(
                    f"Type for {GLOBAL_SILENCE_TIMEOUT_KEY} is wrong, expected number. "
                    f"Got: '{self.global_silence_timeout}'. "
                )

        if not isinstance(self.global_silence_timeout, (float, int)):
            raise RasaException(
                f"Type for {GLOBAL_SILENCE_TIMEOUT_KEY} is wrong, expected number. "
                f"Got: '{type(self.global_silence_timeout)}'. "
            )

        if self.global_silence_timeout <= 0:
            raise RasaException(
                f"Value for {GLOBAL_SILENCE_TIMEOUT_KEY} must be a positive number. "
                f"Got: '{self.global_silence_timeout}'. "
            )

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]) -> InteractionHandlingConfig:
        """Create a InteractionHandlingConfig instance from a dictionary."""
        return cls(
            global_silence_timeout=data.get(
                GLOBAL_SILENCE_TIMEOUT_KEY, GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
            )
            if data is not None
            else GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
        )


class AvailableEndpoints:
    """Collection of configured endpoints."""

    _instance = None

    @classmethod
    def read_endpoints(cls, endpoint_file: str) -> AvailableEndpoints:
        """Read the different endpoints from a yaml file."""
        nlg = read_endpoint_config(endpoint_file, endpoint_type="nlg")
        nlu = read_endpoint_config(endpoint_file, endpoint_type="nlu")
        action = read_endpoint_config(endpoint_file, endpoint_type="action_endpoint")
        model = read_endpoint_config(endpoint_file, endpoint_type="models")
        tracker_store = read_endpoint_config(
            endpoint_file, endpoint_type="tracker_store"
        )
        lock_store = read_endpoint_config(endpoint_file, endpoint_type="lock_store")
        event_broker = read_endpoint_config(endpoint_file, endpoint_type="event_broker")
        vector_store = read_endpoint_config(endpoint_file, endpoint_type="vector_store")
        model_groups = read_property_config_from_endpoints_file(
            endpoint_file, property_name="model_groups"
        )
        privacy = read_property_config_from_endpoints_file(
            endpoint_file, property_name="privacy"
        )

        interaction_handling = InteractionHandlingConfig.from_dict(
            read_property_config_from_endpoints_file(
                endpoint_file, property_name="interaction_handling"
            )
        )

        return cls(
            nlg,
            nlu,
            action,
            model,
            tracker_store,
            lock_store,
            event_broker,
            vector_store,
            model_groups,
            privacy,
            interaction_handling,
        )

    def __init__(
        self,
        nlg: Optional[EndpointConfig] = None,
        nlu: Optional[EndpointConfig] = None,
        action: Optional[EndpointConfig] = None,
        model: Optional[EndpointConfig] = None,
        tracker_store: Optional[EndpointConfig] = None,
        lock_store: Optional[EndpointConfig] = None,
        event_broker: Optional[EndpointConfig] = None,
        vector_store: Optional[EndpointConfig] = None,
        model_groups: Optional[List[Dict[str, Any]]] = None,
        privacy: Optional[Dict[str, Any]] = None,
        interaction_handling: InteractionHandlingConfig = InteractionHandlingConfig(
            global_silence_timeout=GLOBAL_SILENCE_TIMEOUT_DEFAULT_VALUE
        ),
    ) -> None:
        """Create an `AvailableEndpoints` object."""
        self.model = model
        self.action = action
        self.nlu = nlu
        self.nlg = nlg
        self.tracker_store = tracker_store
        self.lock_store = lock_store
        self.event_broker = event_broker
        self.vector_store = vector_store
        self.model_groups = model_groups
        self.privacy = privacy
        self.interaction_handling = interaction_handling

    @classmethod
    def get_instance(
        cls, endpoint_file: Optional[str] = DEFAULT_ENDPOINTS_PATH
    ) -> AvailableEndpoints:
        """Get the singleton instance of AvailableEndpoints."""
        # Ensure that the instance is initialized only once.
        if cls._instance is None:
            cls._instance = cls.read_endpoints(endpoint_file)
        return cls._instance

    @classmethod
    def reset_instance(cls) -> None:
        cls._instance = None



from __future__ import annotations

import importlib.resources
import json
import logging
from copy import deepcopy
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Text,
    Type,
    TypeVar,
    Union,
    cast,
)

import structlog
from pydantic import BaseModel, Field

import rasa.cli.telemetry
import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.core.available_endpoints import AvailableEndpoints
from rasa.shared.constants import (
    CONFIG_NAME_KEY,
    CONFIG_PIPELINE_KEY,
    CONFIG_POLICIES_KEY,
    DEFAULT_PROMPT_PACKAGE_NAME,
    ENDPOINTS_NLG_KEY,
    LLM_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_GROUPS_CONFIG_KEY,
    MODELS_CONFIG_KEY,
    PROMPT_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY,
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG,
    ROUTER_CONFIG_KEY,
)
from rasa.shared.core.events import BotUttered, UserUttered
from rasa.shared.core.slots import BooleanSlot, CategoricalSlot, Slot
from rasa.shared.engine.caching import get_local_cache_location
from rasa.shared.exceptions import (
    FileIOException,
    FileNotFoundException,
    InvalidConfigException,
)
from rasa.shared.providers._configs.azure_openai_client_config import (
    is_azure_openai_config,
)
from rasa.shared.providers._configs.huggingface_local_embedding_client_config import (
    is_huggingface_local_config,
)
from rasa.shared.providers._configs.openai_client_config import is_openai_config
from rasa.shared.providers._configs.self_hosted_llm_client_config import (
    is_self_hosted_config,
)
from rasa.shared.providers.embedding.embedding_client import EmbeddingClient
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.providers.mappings import (
    AZURE_OPENAI_PROVIDER,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER,
    OPENAI_PROVIDER,
    SELF_HOSTED_PROVIDER,
    get_client_config_class_from_provider,
    get_embedding_client_from_provider,
    get_llm_client_from_provider,
)
from rasa.shared.utils.common import all_subclasses
from rasa.shared.utils.constants import LOG_COMPONENT_SOURCE_METHOD_INIT

if TYPE_CHECKING:
    from rasa.core.agent import Agent
    from rasa.shared.core.trackers import DialogueStateTracker


structlogger = structlog.get_logger()

USER = "USER"

AI = "AI"

DEFAULT_OPENAI_GENERATE_MODEL_NAME = "gpt-4o-2024-11-20"

DEFAULT_OPENAI_CHAT_MODEL_NAME = "gpt-4o-2024-11-20"

DEFAULT_ENTERPRISE_SEARCH_POLICY_MODEL_NAME = "gpt-4.1-mini-2025-04-14"

DEFAULT_OPENAI_CHAT_MODEL_NAME_ADVANCED = "gpt-4-0613"

DEFAULT_OPENAI_EMBEDDING_MODEL_NAME = "text-embedding-3-large"

DEFAULT_OPENAI_TEMPERATURE = 0.7

DEFAULT_OPENAI_MAX_GENERATED_TOKENS = 256

DEFAULT_MAX_USER_INPUT_CHARACTERS = 420

DEPLOYMENT_CENTRIC_PROVIDERS = [AZURE_OPENAI_PROVIDER]

# Placeholder messages used in the transcript for
# instances where user input results in an error
ERROR_PLACEHOLDER = {
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG: "[User sent really long message]",
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY: "",
    "default": "[User input triggered an error]",
}

_Factory_F = TypeVar(
    "_Factory_F",
    bound=Callable[[Dict[str, Any], Dict[str, Any]], Union[EmbeddingClient, LLMClient]],
)
_CombineConfigs_F = TypeVar(
    "_CombineConfigs_F",
    bound=Callable[[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
)


class SystemPrompts(BaseModel):
    command_generator: str = Field(
        ..., description="Prompt used by the LLM command generator."
    )
    enterprise_search: str = Field(
        ..., description="Prompt for standard enterprise search requests."
    )
    contextual_response_rephraser: str = Field(
        ..., description="Prompt used for re-phrasing assistant responses."
    )


def _compute_hash_for_cache_from_configs(
    config_x: Dict[str, Any], config_y: Dict[str, Any]
) -> int:
    """Get a unique hash of the default and custom configs."""
    return hash(
        json.dumps(config_x, sort_keys=True) + json.dumps(config_y, sort_keys=True)
    )


def _retrieve_from_cache(
    cache: Dict[int, Any], unique_hash: int, function: Callable, function_kwargs: dict
) -> Any:
    """Retrieve the value from the cache if it exists. If it does not exist, cache it"""
    if unique_hash in cache:
        return cache[unique_hash]
    else:
        return_value = function(**function_kwargs)
        cache[unique_hash] = return_value
        return return_value


def _cache_factory(function: _Factory_F) -> _Factory_F:
    """Memoize the factory methods based on the arguments."""
    cache: Dict[int, Union[EmbeddingClient, LLMClient]] = {}

    @wraps(function)
    def factory_method_wrapper(
        config_x: Dict[str, Any], config_y: Dict[str, Any]
    ) -> Union[EmbeddingClient, LLMClient]:
        # Get a unique hash of the default and custom configs.
        unique_hash = _compute_hash_for_cache_from_configs(config_x, config_y)
        return _retrieve_from_cache(
            cache=cache,
            unique_hash=unique_hash,
            function=function,
            function_kwargs={"custom_config": config_x, "default_config": config_y},
        )

    def clear_cache() -> None:
        cache.clear()
        structlogger.debug(
            "Cleared cache for factory method",
            function_name=function.__name__,
        )

    setattr(factory_method_wrapper, "clear_cache", clear_cache)
    return cast(_Factory_F, factory_method_wrapper)


def _cache_combine_custom_and_default_configs(
    function: _CombineConfigs_F,
) -> _CombineConfigs_F:
    """Memoize the combine_custom_and_default_config method based on the arguments."""
    cache: Dict[int, dict] = {}

    @wraps(function)
    def combine_configs_wrapper(
        config_x: Dict[str, Any], config_y: Dict[str, Any]
    ) -> dict:
        # Get a unique hash of the default and custom configs.
        unique_hash = _compute_hash_for_cache_from_configs(config_x, config_y)
        return _retrieve_from_cache(
            cache=cache,
            unique_hash=unique_hash,
            function=function,
            function_kwargs={"custom_config": config_x, "default_config": config_y},
        )

    def clear_cache() -> None:
        cache.clear()
        structlogger.debug(
            "Cleared cache for combine_custom_and_default_config method",
            function_name=function.__name__,
        )

    setattr(combine_configs_wrapper, "clear_cache", clear_cache)
    return cast(_CombineConfigs_F, combine_configs_wrapper)


def tracker_as_readable_transcript(
    tracker: "DialogueStateTracker",
    human_prefix: str = USER,
    ai_prefix: str = AI,
    max_turns: Optional[int] = 20,
    turns_wrapper: Optional[Callable[[List[str]], List[str]]] = None,
) -> str:
    """Creates a readable dialogue from a tracker.

    Args:
        tracker: the tracker to convert
        human_prefix: the prefix to use for human utterances
        ai_prefix: the prefix to use for ai utterances
        max_turns: the maximum number of turns to include in the transcript
        turns_wrapper: optional function to wrap the turns in a custom way

    Example:
        >>> tracker = Tracker(
        ...     sender_id="test",
        ...     slots=[],
        ...     events=[
        ...         UserUttered("hello"),
        ...         BotUttered("hi"),
        ...     ],
        ... )
        >>> tracker_as_readable_transcript(tracker)
        USER: hello
        AI: hi

    Returns:
    A string representing the transcript of the tracker
    """
    transcript = []

    # using `applied_events` rather than `events` means that only events after the
    # most recent `Restart` or `SessionStarted` are included in the transcript
    for event in tracker.applied_events():
        if isinstance(event, UserUttered):
            if event.has_triggered_error:
                first_error = event.error_commands[0]
                error_type = first_error.get("error_type")
                message = ERROR_PLACEHOLDER.get(
                    error_type, ERROR_PLACEHOLDER["default"]
                )
            else:
                message = sanitize_message_for_prompt(event.text)
            transcript.append(f"{human_prefix}: {message}")

        elif isinstance(event, BotUttered):
            transcript.append(f"{ai_prefix}: {sanitize_message_for_prompt(event.text)}")

    # turns_wrapper to count multiple utterances by bot/user as single turn
    if turns_wrapper:
        transcript = turns_wrapper(transcript)
    # otherwise, just take the last `max_turns` lines of the transcript
    transcript = transcript[-max_turns if max_turns is not None else None :]

    return "\n".join(transcript)


def sanitize_message_for_prompt(text: Optional[str]) -> str:
    """Removes new lines from a string.

    Args:
        text: the text to sanitize

    Returns:
    A string with new lines removed.
    """
    return text.replace("\n", " ") if text else ""


@_cache_combine_custom_and_default_configs
def combine_custom_and_default_config(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> Dict[Text, Any]:
    """Merges the given model configuration with the default configuration.

    This method supports both single model configurations and model group configurations
    (configs that have the `models` key).

    If `custom_config` is a single model configuration, it merges `custom_config` with
    `default_config`, which is also a single model configuration.

    If `custom_config` is a model group configuration (contains the `models` key), it
    applies the merging process to each model configuration within the group
    individually, merging each with the `default_config`.

    Note that `default_config` is always a single model configuration.

    The method ensures that the provider is set and all deprecated keys are resolved,
    resulting in a valid client configuration.

    Args:
        custom_config: The custom configuration containing values to overwrite defaults.
            Can be a single model configuration or a model group configuration with a
            `models` key.
        default_config: The default configuration, which is a single model
            configuration.

    Returns:
        The merged configuration, either a single model configuration or a model group
        configuration with merged models.
    """
    if custom_config and MODELS_CONFIG_KEY in custom_config:
        return _combine_model_groups_configs_with_default_config(
            custom_config, default_config
        )
    else:
        return _combine_single_model_configs(custom_config, default_config)


def _combine_model_groups_configs_with_default_config(
    model_group_config: Dict[str, Any], default_config: Dict[str, Any]
) -> Dict[Text, Any]:
    """Merges each model configuration within a model group with the default
    configuration.

    This method processes model group configurations by applying the merging process to
    each model configuration within the group individually.

    Args:
        model_group_config: The model group configuration containing a list of model
            configurations under the `models` key.
        default_config: The default configuration for a single model.

    Returns:
        The merged model group configuration with each model configuration merged
        with the default configuration.
    """
    model_group_config = deepcopy(model_group_config)
    model_group_config_combined_with_defaults = [
        _combine_single_model_configs(model_config, default_config)
        for model_config in model_group_config[MODELS_CONFIG_KEY]
    ]
    # Update the custom models config with the combined config.
    model_group_config[MODELS_CONFIG_KEY] = model_group_config_combined_with_defaults
    return model_group_config


@_cache_combine_custom_and_default_configs
def _combine_single_model_configs(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> Dict[Text, Any]:
    """Merges the given model config with the default config.

    This method guarantees that the provider is set and all the deprecated keys are
    resolved. Hence, produces only a valid client config.

    Only uses the default configuration arguments, if the type set in the
    custom config matches the type in the default config. Otherwise, only
    the custom config is used.

    Args:
        custom_config: The custom config containing values to overwrite defaults
        default_config: The default config.

    Returns:
        The merged config.
    """
    if custom_config is None:
        return default_config.copy()

    # Get the provider from the custom config.
    custom_config_provider = get_provider_from_config(custom_config)
    # We expect the provider to be set in the default configs of all Rasa components.
    default_config_provider = default_config[PROVIDER_CONFIG_KEY]

    if (
        custom_config_provider is not None
        and custom_config_provider != default_config_provider
    ):
        # Get the provider-specific config class
        client_config_clazz = get_client_config_class_from_provider(
            custom_config_provider
        )
        # Checks for deprecated keys, resolves aliases and returns a valid config.
        # This is done to ensure that the custom config is valid.
        return client_config_clazz.from_dict(deepcopy(custom_config)).to_dict()

    # If the provider is the same in both configs
    # OR provider is not specified in the custom config
    # perform MERGE by overriding the default config keys and values
    # with custom config keys and values.
    merged_config = {**deepcopy(default_config), **deepcopy(custom_config)}
    # Check for deprecated keys, resolve aliases and return a valid config.
    # This is done to ensure that the merged config is valid.
    default_config_clazz = get_client_config_class_from_provider(
        default_config_provider
    )
    return default_config_clazz.from_dict(merged_config).to_dict()


def get_provider_from_config(config: dict) -> Optional[str]:
    """Try to get the provider from the passed llm/embeddings configuration.
    If no provider can be found, return None.
    """
    if not config:
        return None
    if is_self_hosted_config(config):
        return SELF_HOSTED_PROVIDER
    elif is_azure_openai_config(config):
        return AZURE_OPENAI_PROVIDER
    elif is_openai_config(config):
        return OPENAI_PROVIDER
    elif is_huggingface_local_config(config):
        return HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER
    else:
        return config.get(PROVIDER_CONFIG_KEY)


def ensure_cache() -> None:
    """Ensures that the cache is initialized."""
    import litellm

    # Ensure the cache directory exists
    cache_location = get_local_cache_location() / "rasa-llm-cache"
    cache_location.mkdir(parents=True, exist_ok=True)

    # Set diskcache as a caching option
    litellm.cache = litellm.Cache(type="disk", disk_cache_dir=cache_location)


@_cache_factory
def llm_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> LLMClient:
    """Creates an LLM from the given config.

    If the config is using the old syntax, e.g. defining the llm client directly in
    config.yaml, then standalone client is initialised (no routing).

    If the config uses the using the new, model group syntax, defined in the
    endpoints.yml, then router client is initialised if there are more than one model
    within the group.

    Examples:
    The config below will result in a standalone client:
    ```
    {
       "provider": "openai",
       "model": "gpt-4",
       "timeout": 10,
       "num_retries": 3,
    }
    ```

    The config below will also result in a standalone client:
    ```
    {
        "id": "model-group-id",
        "models": [
            {"provider": "openai", "model": "gpt-4", "api_key": "test"},
        ],
    }
    ```

    The config below will result in a router client:
    ```
    {
        "id": "test-model-group-id",
        "models": [
            {"provider": "openai", "model": "gpt-4", "api_key": "test"},
            {
                "provider": "azure",
                "deployment": "test-deployment",
                "api_key": "test",
                "api_base": "test-api-base",
            },
        ],
        "router": {"routing_strategy": "test"},
    }
    ```

    Args:
        custom_config: The custom config  containing values to overwrite defaults.
        default_config: The default config.

    Returns:
        Instantiated client based on the configuration.
    """
    if custom_config:
        if ROUTER_CONFIG_KEY in custom_config:
            return llm_router_factory(custom_config, default_config)
        if MODELS_CONFIG_KEY in custom_config:
            return llm_client_factory(
                custom_config[MODELS_CONFIG_KEY][0], default_config
            )
    return llm_client_factory(custom_config, default_config)


def llm_router_factory(
    router_config: Dict[str, Any], default_model_config: Dict[str, Any], **kwargs: Any
) -> LLMClient:
    """Creates an LLM Router using the provided configurations.

    This function initializes an LLM Router based on the given router configuration,
    which includes multiple model configurations. For each model specified in the router
    configuration, any missing parameters are supplemented using the default model
    configuration.

    Args:
        router_config: The full router configuration containing multiple model
            configurations. Each model's configuration can override parameters from the
            default model configuration.
        default_model_config: The default configuration parameters for a single model.
            These defaults are used to fill in any missing parameters in each model's
            configuration within the router.

    Returns:
        An instance that conforms to both `LLMClient` and `RouterClient` protocols
        representing the configured LLM Router.
    """
    from rasa.shared.providers.llm.litellm_router_llm_client import (
        LiteLLMRouterLLMClient,
    )

    combined_config = _combine_model_groups_configs_with_default_config(
        router_config, default_model_config
    )
    return LiteLLMRouterLLMClient.from_config(combined_config)


def llm_client_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> LLMClient:
    """Creates an LLM from the given config.

    Args:
        custom_config: The custom config  containing values to overwrite defaults
        default_config: The default config.

    Returns:
        Instantiated LLM based on the configuration.
    """
    config = combine_custom_and_default_config(deepcopy(custom_config), default_config)

    ensure_cache()

    client_clazz: Type[LLMClient] = get_llm_client_from_provider(
        config[PROVIDER_CONFIG_KEY]
    )
    client = client_clazz.from_config(config)
    return client


@_cache_factory
def embedder_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> EmbeddingClient:
    """Creates an embedding client from the given config.

    If the config is using the old syntax, e.g. defining the llm client directly in
    config.yaml, then standalone client is initialised (no routing).

    If the config uses the using the new, model group syntax, defined in the
    endpoints.yml, then router client is initialised if there are more than one model
    within the group and the router is defined.

    Examples:
    The config below will result in a standalone client:
    ```
    {
       "provider": "openai",
       "model": "text-embedding-3-large",
       "timeout": 10,
       "num_retries": 3,
    }
    ```

    The config below will also result in a standalone client:
    ```
    {
        "id": "model-group-id",
        "models": [
            {
                "provider": "openai",
                "model": "test-embedding-3-large",
                "api_key": "test"
            },
        ],
    }
    ```

    The config below will result in a router client:
    ```
    {
        "id": "test-model-group-id",
        "models": [
            {"provider": "openai", "model": "gpt-4", "api_key": "test"},
            {
                "provider": "azure",
                "deployment": "test-deployment",
                "api_key": "test",
                "api_base": "test-api-base",
            },
        ],
        "router": {"routing_strategy": "test"},
    }
    ```

    Args:
        custom_config: The custom config  containing values to overwrite defaults.
        default_config: The default config.

    Returns:
        Instantiated client based on the configuration.
    """
    if custom_config:
        if ROUTER_CONFIG_KEY in custom_config:
            return embedder_router_factory(custom_config, default_config)
        if MODELS_CONFIG_KEY in custom_config:
            return embedder_client_factory(
                custom_config[MODELS_CONFIG_KEY][0], default_config
            )
    return embedder_client_factory(custom_config, default_config)


def embedder_router_factory(
    router_config: Dict[str, Any], default_model_config: Dict[str, Any], **kwargs: Any
) -> EmbeddingClient:
    """Creates an Embedder Router using the provided configurations.

    This function initializes an Embedder Router based on the given router
    configuration, which includes multiple model configurations. For each model
    specified in the router configuration, any missing parameters are supplemented using
    the default model configuration.

    Args:
        router_config: The full router configuration containing multiple model
            configurations. Each model's configuration can override parameters from the
            default model configuration.
        default_model_config: The default configuration parameters for a single model.
            These defaults are used to fill in any missing parameters in each model's
            configuration within the router.

    Returns:
        An instance that conforms to both `EmbeddingClient` and `RouterClient` protocols
        representing the configured Embedding Router.
    """
    from rasa.shared.providers.embedding.litellm_router_embedding_client import (
        LiteLLMRouterEmbeddingClient,
    )

    combined_config = _combine_model_groups_configs_with_default_config(
        router_config, default_model_config
    )

    return LiteLLMRouterEmbeddingClient.from_config(combined_config)


def embedder_client_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> EmbeddingClient:
    """Creates an Embedder from the given config.

    Args:
        custom_config: The custom config containing values to overwrite defaults
        default_config: The default config.


    Returns:
        Instantiated Embedder based on the configuration.
    """
    config = combine_custom_and_default_config(deepcopy(custom_config), default_config)

    ensure_cache()

    client_clazz: Type[EmbeddingClient] = get_embedding_client_from_provider(
        config[PROVIDER_CONFIG_KEY]
    )
    client = client_clazz.from_config(config)
    return client


def get_prompt_template(
    jinja_file_path: Optional[Text],
    default_prompt_template: Text,
    *,
    log_source_component: Optional[Text] = None,
    log_source_method: Optional[Literal["init", "fingerprint_addon"]] = None,
) -> Text:
    """Returns the jinja template.

    Args:
        jinja_file_path: The path to the jinja template file. If not provided, the
            default template will be used.
        default_prompt_template: The fallback prompt template to use if no file is
            found or specified.
        log_source_component: The name of the component emitting the log, used to
            identify the source in structured logging.
        log_source_method: The name of the method or function emitting the log for
            better traceability.

    Returns:
        The prompt template.
    """
    try:
        if jinja_file_path is not None:
            prompt_template = rasa.shared.utils.io.read_file(jinja_file_path)

            log_level = (
                logging.INFO
                if log_source_method == LOG_COMPONENT_SOURCE_METHOD_INIT
                else logging.DEBUG
            )

            structlogger.log(
                log_level,
                "utils.llm.get_prompt_template"
                ".custom_prompt_template_read_successfully",
                event_info=(
                    f"Custom prompt template read successfully from "
                    f"`{jinja_file_path}`."
                ),
                prompt_file_path=jinja_file_path,
                log_source_component=log_source_component,
                log_source_method=log_source_method,
            )
            return prompt_template
    except (FileIOException, FileNotFoundException):
        structlogger.warning(
            "utils.llm.get_prompt_template" ".failed_to_read_custom_prompt_template",
            event_info=(
                "Failed to read custom prompt template. Using default template instead."
            ),
            log_source_component=log_source_component,
            log_source_method=log_source_method,
        )
    return default_prompt_template


def get_default_prompt_template_based_on_model(
    llm_config: Dict[str, Any],
    model_prompt_mapping: Dict[str, Any],
    default_prompt_path: str,
    fallback_prompt_path: str,
    *,
    log_source_component: Optional[Text] = None,
    log_source_method: Optional[Literal["init", "fingerprint_addon"]] = None,
) -> Text:
    """Returns the default prompt template based on the model name.

    Args:
        llm_config: The model config.
        model_prompt_mapping: The model name -> prompt template mapping.
        default_prompt_path: The path to the default prompt template for the component.
        fallback_prompt_path: The fallback prompt path for all other models that do not
            have a mapping in the model_prompt_mapping.
        log_source_component: The name of the component emitting the log, used to
            identify the source in structured logging.
        log_source_method: The name of the method or function emitting the log for
            better traceability.

    Returns:
        The default prompt template.
    """
    # Extract the provider and model name information from the configuration
    _llm_config = deepcopy(llm_config)
    if MODELS_CONFIG_KEY in _llm_config:
        _llm_config = _llm_config[MODELS_CONFIG_KEY][0]
    provider = _llm_config.get(PROVIDER_CONFIG_KEY)
    model = _llm_config.get(MODEL_CONFIG_KEY)

    # If the model is not defined, we default to the default prompt template.
    if not model:
        structlogger.debug(
            "utils.llm.get_default_prompt_template_based_on_model"
            ".using_default_prompt_template",
            event_info=(
                f"Model not defined in the config. Default prompt template read from"
                f" - `{default_prompt_path}`."
            ),
            default_prompt_path=default_prompt_path,
            log_source_component=log_source_component,
            log_source_method=log_source_method,
        )
        return importlib.resources.read_text(
            DEFAULT_PROMPT_PACKAGE_NAME, default_prompt_path
        )

    full_model_name = model if provider and provider in model else f"{provider}/{model}"

    # If the model is found in the mapping, we use the model-specific prompt
    # template.
    if prompt_file_path := model_prompt_mapping.get(full_model_name):
        structlogger.debug(
            "utils.llm.get_default_prompt_template_based_on_model"
            ".using_model_specific_prompt_template",
            event_info=(
                f"Using model-specific default prompt template. Default prompt "
                f"template read from - `{prompt_file_path}`."
            ),
            default_prompt_path=prompt_file_path,
            model_name=full_model_name,
            log_source_component=log_source_component,
            log_source_method=log_source_method,
        )
        return importlib.resources.read_text(
            DEFAULT_PROMPT_PACKAGE_NAME, prompt_file_path
        )

    # If the model is not found in the mapping, we default to the fallback prompt
    # template.
    structlogger.debug(
        "utils.llm.get_default_prompt_template_based_on_model"
        ".using_fallback_prompt_template",
        event_info=(
            f"Model not found in the model prompt mapping. Fallback prompt template "
            f"read from - `{fallback_prompt_path}`."
        ),
        fallback_prompt_path=fallback_prompt_path,
        model_name=full_model_name,
        log_source_component=log_source_component,
        log_source_method=log_source_method,
    )
    return importlib.resources.read_text(
        DEFAULT_PROMPT_PACKAGE_NAME, fallback_prompt_path
    )


def allowed_values_for_slot(slot: Slot) -> Union[str, None]:
    """Get the allowed values for a slot."""
    if isinstance(slot, BooleanSlot):
        return str([True, False])
    if isinstance(slot, CategoricalSlot):
        return str([v for v in slot.values if v != "__other__"])
    else:
        return None


def resolve_model_client_config(
    model_config: Optional[Dict[str, Any]],
    component_name: Optional[str] = None,
    model_groups: Optional[List[Dict[str, Any]]] = None,
) -> Optional[Dict[str, Any]]:
    """Resolve the model group in the model config.

    1. If the config is pointing to a model group, the corresponding model group
    of the endpoints.yml is returned.
    2. If the config is using the old syntax, e.g. defining the llm
    directly in config.yml, the config is returned as is.
    3. If the config is already resolved, return it as is.

    Args:
        model_config: The model config to be resolved.
        component_name: The name of the component.
        component_name: The method of the component.
        model_groups: Model groups from endpoints.yml.

    Returns:
        The resolved llm config.
    """

    def _raise_invalid_config_exception(reason: str) -> None:
        """Helper function to raise InvalidConfigException with a formatted message."""
        if component_name:
            message = (
                f"Could not resolve model group '{model_group_id}'"
                f" for component '{component_name}'."
            )
        else:
            message = f"Could not resolve model group '{model_group_id}'."
        message += f" {reason}"
        raise InvalidConfigException(message)

    if model_config is None:
        return None

    # Config is already resolved or defines a client without model groups
    if MODEL_GROUP_CONFIG_KEY not in model_config:
        return model_config

    model_group_id = model_config.get(MODEL_GROUP_CONFIG_KEY)

    # If `model_groups` is provided, use it to initialise `AvailableEndpoints`,
    # since `get_instance()` reads from the local endpoints file instead.
    if model_groups:
        endpoints = AvailableEndpoints(model_groups=model_groups)
    else:
        endpoints = AvailableEndpoints.get_instance()
    if endpoints.model_groups is None:
        _raise_invalid_config_exception(
            reason=(
                "No model group with that id found in endpoints.yml. "
                "Please make sure to define the model group."
            )
        )

    copy_model_groups = deepcopy(endpoints.model_groups)
    model_group = [
        model_group
        for model_group in copy_model_groups  # type: ignore[union-attr]
        if model_group.get(MODEL_GROUP_ID_CONFIG_KEY) == model_group_id
    ]

    if len(model_group) == 0:
        _raise_invalid_config_exception(
            reason=(
                "No model group with that id found in endpoints.yml. "
                "Please make sure to define the model group."
            )
        )
    if len(model_group) > 1:
        _raise_invalid_config_exception(
            reason=(
                "Multiple model groups with that id found in endpoints.yml. "
                "Please make sure to define the model group just once."
            )
        )

    return model_group[0]


def generate_sender_id(test_case_name: str) -> str:
    # add timestamp suffix to ensure sender_id is unique
    return f"{test_case_name}_{datetime.now()}"


async def create_tracker_for_user_step(
    step_sender_id: str,
    agent: "Agent",
    test_case_tracker: "DialogueStateTracker",
    index_user_uttered_event: int,
) -> None:
    """Creates a tracker for the user step."""
    tracker = test_case_tracker.copy()
    # modify the sender id so that the original tracker is not overwritten
    tracker.sender_id = step_sender_id

    if tracker.events:
        # get the timestamp of the event just before the user uttered event
        timestamp = tracker.events[index_user_uttered_event - 1].timestamp
        # revert the tracker to the event just before the user uttered event
        tracker = tracker.travel_back_in_time(timestamp)

    # store the tracker with the unique sender id
    await agent.tracker_store.save(tracker)


def check_prompt_config_keys_and_warn_if_deprecated(
    config: dict, component_source: str
) -> None:
    """Checks and warns about deprecated config parameters."""
    if PROMPT_CONFIG_KEY in config and PROMPT_TEMPLATE_CONFIG_KEY in config:
        structlogger.warning(
            f"{component_source}.init"
            ".both_deprecated_and_non_deprecated_config_keys_used_at_the_same_time",
            event_info=(
                f"Both '{PROMPT_CONFIG_KEY}' and '{PROMPT_TEMPLATE_CONFIG_KEY}' "
                f"are present in the config. '{PROMPT_CONFIG_KEY}' will be ignored "
                f"in favor of {PROMPT_TEMPLATE_CONFIG_KEY}."
            ),
        )

    # 'prompt' config key is deprecated in favor of 'prompt_template'
    if PROMPT_CONFIG_KEY in config:
        structlogger.warning(
            f"{component_source}.init.deprecated_config_key",
            event_info=(
                f"The config parameter '{PROMPT_CONFIG_KEY}' is deprecated "
                "and will be removed in Rasa 4.0.0. "
                f"Please use the config parameter '{PROMPT_TEMPLATE_CONFIG_KEY}'"
                f" instead. "
            ),
        )


def _get_llm_command_generator_config(
    config: Dict[Text, Any],
) -> Optional[Dict[Text, Any]]:
    """Get the llm command generator config from config.yml.

    Args:
        config: The config.yml file data.

    Returns:
        The llm command generator config.
    """
    from rasa.dialogue_understanding.generator import LLMBasedCommandGenerator

    # Collect all LLM based Command Generator class names.
    command_generator_subclasses = all_subclasses(LLMBasedCommandGenerator)
    command_generator_class_names = [
        command_generator.__name__ for command_generator in command_generator_subclasses
    ]

    # Read the LLM config of the Command Generator from the config.yml file.
    pipelines = config.get(CONFIG_PIPELINE_KEY, [])
    for pipeline in pipelines:
        if pipeline.get(CONFIG_NAME_KEY) in command_generator_class_names:
            return pipeline.get(LLM_CONFIG_KEY)

    return None


def _get_command_generator_prompt(
    config: Dict[Text, Any], endpoints: Dict[Text, Any]
) -> Text:
    """Get the command generator prompt based on the config."""
    from rasa.dialogue_understanding.generator.single_step.compact_llm_command_generator import (  # noqa: E501
        DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
        FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
        MODEL_PROMPT_MAPPER,
    )

    model_config = _get_llm_command_generator_config(config)
    llm_config = resolve_model_client_config(
        model_config=model_config,
        model_groups=endpoints.get(MODEL_GROUPS_CONFIG_KEY),
    )
    return get_default_prompt_template_based_on_model(
        llm_config=llm_config or {},
        model_prompt_mapping=MODEL_PROMPT_MAPPER,
        default_prompt_path=DEFAULT_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
        fallback_prompt_path=FALLBACK_COMMAND_PROMPT_TEMPLATE_FILE_NAME,
    )


def _get_enterprise_search_prompt(config: Dict[Text, Any]) -> Text:
    """Get the enterprise search prompt based on the config."""
    from rasa.core.policies.enterprise_search_policy import EnterpriseSearchPolicy

    def get_enterprise_search_config() -> Dict[Text, Any]:
        policies = config.get(CONFIG_POLICIES_KEY, [])
        for policy in policies:
            if policy.get(CONFIG_NAME_KEY) == EnterpriseSearchPolicy.__name__:
                return policy

        return {}

    enterprise_search_config = get_enterprise_search_config()
    return EnterpriseSearchPolicy.get_system_default_prompt_based_on_config(
        enterprise_search_config
    )


def get_system_default_prompts(
    config: Dict[Text, Any], endpoints: Dict[Text, Any]
) -> SystemPrompts:
    """Returns the system default prompts for the component.

    Args:
        config: The config.yml file data.
        endpoints: The endpoints.yml file data.

    Returns:
        SystemPrompts: A Pydantic model containing all default prompts.
    """
    from rasa.core.nlg.contextual_response_rephraser import (
        DEFAULT_RESPONSE_VARIATION_PROMPT_TEMPLATE,
    )

    return SystemPrompts(
        command_generator=_get_command_generator_prompt(config, endpoints),
        enterprise_search=_get_enterprise_search_prompt(config),
        contextual_response_rephraser=DEFAULT_RESPONSE_VARIATION_PROMPT_TEMPLATE,
    )


def collect_custom_prompts(
    config: Dict[Text, Any],
    endpoints: Dict[Text, Any],
    project_root: Optional[Path] = None,
) -> Dict[Text, Text]:
    """Collects custom prompts from the project configuration and endpoints.

    Args:
        config: The configuration dictionary of the project.
        endpoints: The endpoints configuration dictionary.
        project_root: The root directory of the project.

    Returns:
        A dictionary containing custom prompts.
        The keys are:
            - 'contextual_response_rephraser'
            - 'command_generator'
            - 'enterprise_search'
    """
    from rasa.core.policies.enterprise_search_policy import EnterpriseSearchPolicy
    from rasa.dialogue_understanding.generator.llm_based_command_generator import (
        LLMBasedCommandGenerator,
    )
    from rasa.studio.prompts import (
        COMMAND_GENERATOR_NAME,
        CONTEXTUAL_RESPONSE_REPHRASER_NAME,
        ENTERPRISE_SEARCH_NAME,
    )

    prompts: Dict[Text, Text] = {}
    project_root = project_root or Path(".").resolve()

    def _read_prompt(root: Path, path_in_yaml: Text) -> Optional[Text]:
        if not path_in_yaml:
            return None

        prompt_path = (
            (root / path_in_yaml).resolve()
            if not Path(path_in_yaml).is_absolute()
            else Path(path_in_yaml)
        )
        if prompt_path.exists():
            return prompt_path.read_text(encoding="utf-8")

        structlogger.warning(
            "utils.llm.collect_custom_prompts.prompt_not_found",
            event_info=(f"Prompt file '{prompt_path}' not found. "),
            prompt_path=prompt_path,
            project_root=root,
        )
        return None

    # contextual_response_rephraser
    nlg_conf = endpoints.get(ENDPOINTS_NLG_KEY) or {}
    if prompt_text := _read_prompt(project_root, nlg_conf.get(PROMPT_CONFIG_KEY)):
        prompts[CONTEXTUAL_RESPONSE_REPHRASER_NAME] = prompt_text

    # command_generator
    command_generator_classes = {
        cls.__name__ for cls in all_subclasses(LLMBasedCommandGenerator)
    }
    for component in config.get(CONFIG_PIPELINE_KEY, []):
        if component.get(CONFIG_NAME_KEY) in command_generator_classes:
            if prompt_text := _read_prompt(
                project_root, component.get(PROMPT_TEMPLATE_CONFIG_KEY)
            ):
                prompts[COMMAND_GENERATOR_NAME] = prompt_text
                break

    # enterprise_search
    for policy in config.get(CONFIG_POLICIES_KEY, []):
        if policy.get(CONFIG_NAME_KEY) == EnterpriseSearchPolicy.__name__:
            if prompt_text := _read_prompt(project_root, policy.get(PROMPT_CONFIG_KEY)):
                prompts[ENTERPRISE_SEARCH_NAME] = prompt_text
            break

    return prompts



import json


def to_json_escaped_string(s: str) -> str:
    """
    Serializes the input string using JSON encoding to escape special characters
    like newlines (\\n), tabs (\\t), and quotes.
    """
    return json.dumps(s, ensure_ascii=False)
